// ----------------------------------------------------------------------------
//
// rt_stack.i
// RT++ stack information
//
// $Id: rt_stack.i,v 1.6 1996/04/02 07:33:35 schreine Exp $
// (C) 1996 Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
//
// ----------------------------------------------------------------------------
#ifndef RT_STACK_I
#define RT_STACK_I

// stack direction must be specified in rt_machine.h
#if !defined(RT_GROW_UP) && !defined(RT_GROW_DOWN)
#error no stack direction specified
#endif
#if defined(RT_GROW_UP) && defined(RT_GROW_DOWN)
#error contradicting stack directions specified
#endif

// ----------------------------------------------------------------------------
//
// set stack to empty
//
// ----------------------------------------------------------------------------
inline void
Stack::init(void)
{
  shadow = 0;
  gc = 0;
}

// ----------------------------------------------------------------------------
//
// return true if stack is empty
//
// ----------------------------------------------------------------------------
inline int
Stack::null(void)
{
  return(shadow == 0);
}

// ----------------------------------------------------------------------------
//
// align stack downwards as required for shadow stack
//
// ----------------------------------------------------------------------------
inline Stack::Reference *
Stack::aligndown(void *s) 
{
  static const int align = sizeof(Reference)-1;
  return((Reference *)(((long int)(s)) & ~align));
}

// ----------------------------------------------------------------------------
//
// align stack upwards as required for shadow stack
//
// ----------------------------------------------------------------------------
inline Stack::Reference *
Stack::alignup(void *s) 
{
  static const int align = sizeof(Reference)-1;
  return((Reference *)((((long int)(s)) + align) & ~align));
}

// ----------------------------------------------------------------------------
//
// initialize stack where <stack> is the address of the real stack with <size>
//
// ----------------------------------------------------------------------------
inline void
Stack::init(void *stack)
{
#ifdef RT_GROW_UP
  shadow = aligndown(stack);   // shadow stack grows downwards
#else
  shadow = alignup(stack);     // shadow stack grows upwards
#endif
  top = shadow;                // top of stack
}

// ----------------------------------------------------------------------------
//
// push reference address and type on shadow stack
//
// ----------------------------------------------------------------------------
inline void 
Stack::push(void *_addr, void(*_fun)(void*))
{
  top->addr = _addr;
  top->fun = _fun;
#ifdef RT_GROW_UP
  top--;
#else
  top++;
#endif
#ifdef RT_STACK_CHECK
  Reference a;
  void *p = &a.fun;
#ifdef RT_GROW_UP
  if (p >= (void*)top) rt_abort("stack overflow\n");
#else
  if (p <= (void*)top) rt_abort("stack overflow\n");
#endif
#endif
}

// ----------------------------------------------------------------------------
//
// pop reference address <r> from shadow stack
// <r> need not be on the top of the shadow stack because
// return objects may be allocated after local function variables
//
// ----------------------------------------------------------------------------
inline void 
Stack::pop(void *_addr)
{
#ifdef RT_GROW_UP
  Reference *_t = top+1;
#else
  Reference *_t = top-1;
#endif
  if (_t->addr != _addr)
    {
      Reference *t = _t;
#ifdef RT_GROW_UP
      do { t++; } while (t->addr != _addr);
#else
      do { t--; } while (t->addr != _addr);
#endif
      *t = *_t;
    }
  top = _t;
}

// ----------------------------------------------------------------------------
//
// return upper bound (inclusive) of shadow stack
//
// ----------------------------------------------------------------------------
inline Stack::Reference *
Stack::upper(void)
{
#ifdef RT_GROW_UP
  return(shadow);
#else
  return(top-1);
#endif
}

// ----------------------------------------------------------------------------
//
// return lower bound (exclusive) of shadow stack
//
// ----------------------------------------------------------------------------
inline Stack::Reference *
Stack::lower(void)
{
#ifdef RT_GROW_UP
  return(top);
#else
  return(shadow-1);
#endif
}

// ----------------------------------------------------------------------------
//
// mark all referenced heap blocks
//
// ----------------------------------------------------------------------------
inline void
Stack::mark(void)
{
  Reference *up = upper();
  Reference *low = lower();
  while (up > low)
    {
      void *addr = up->addr;
      void (*fun)(void*) = up->fun;
      fun(addr);
      up--;
    }
}

// ----------------------------------------------------------------------------
//
// check for stack overflow
//
// ----------------------------------------------------------------------------
inline void
Stack::check(void)
{
#ifdef RT_QT
  int a;         // allocate dummy object on stack
  void *p = &a;  // get address of stack object
#ifdef RT_GROW_UP
  if (p+1000 > (void*)top) rt_abort("stack overflow");
#else
  if ((void*)top+1000 > p) rt_abort("stack overflow");
#endif
#endif
}

#endif
// ----------------------------------------------------------------------------
// end of rt_stack.i
// ----------------------------------------------------------------------------
// Local Variables: ***
// mode: C++ ***
// End: ***
