// ----------------------------------------------------------------------------
//
// rt_tbag.h
// RT++ thread bags
//
// $Id: rt_tbag.h,v 1.20 1996/03/22 08:40:35 schreine Exp $
// (C) 1996 Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
//
// ----------------------------------------------------------------------------
#ifndef RT_TBAG_H
#define RT_TBAG_H

// ----------------------------------------------------------------------------
//
// 1. ThreadBagCore:   untyped bag
// 2. ThreadBagBase:   typed bag
// 3. ThreadBag:       pointer to typed bag
// 4. ResBagThread:    bag thread typed in result
// 5. ArgBagThread<i>: bag thread typed in arguments
// 6. ThreadBag<i>:    pointer to typed bag threads
//
// The user only sees <ThreadBag> and <ThreadBag<i>>!
//
// ----------------------------------------------------------------------------

// ----------------------------------------------------------------------------
//
// core of a bag thread
//
// ----------------------------------------------------------------------------
class CoreBagThread: public CoreThread
{
  public:
  CoreBagThread *bnext; // link to next thread in bag
  CoreBagThread *bprev; // link to previous thread in bag
};

// ----------------------------------------------------------------------------
//
// thread list
//
// ----------------------------------------------------------------------------
class ThreadList
{
  public:
  CoreThread *list; // first thread list
  ThreadList *next; // link to next thread list
};

// ----------------------------------------------------------------------------
//
// a bag of threads
//
// ----------------------------------------------------------------------------
class ThreadBagCore
{
  protected:

  Mutex mutex;                 // synchronization on waiters
  int closed;                  // is bag closed?
  int rest;                    // number of remaining threads after closed
  int stopped;                 // is bag stopped?
  int threads;                 // number of threads
  int results;                 // number of thread results available
  CoreBagThread *tlist;        // list of executing threads
  ThreadList *wait_list;       // lists of waiting threads before closing
  ThreadList *wait_tail;       // tail of waiting lists
  CoreThread** wait_array;     // array of waiting threads after closing
  CoreThread *wait_ahead;      // list of ahead waiting threads
  CoreThread *wait_size;       // list of size waiting threads

  void wait(int n);            // wait until result available
  int prepare(void);           // prepare for delivery of next result
  void awake(CoreThread*);     // awake thread list
  void awake(void);            // awake blocked threads
  void deliver(CoreBagThread*);// deliver results
  void close(int k);           // close thread list with <k> more threads
  void scan(void);             // scan of local data structures

  public:

  void init(void);             // initialize empty bag
  void add(CoreBagThread*);    // add thread to bag and return thread number
  int size(void);              // number of threads in bag
  void stop(void);             // stop threads associated to bag

};

// ----------------------------------------------------------------------------
//
// a list node with element of type <T>
//
// ----------------------------------------------------------------------------

template <class T>
class ResultList
{
  public:
  T head;              // the head element
  ResultList<T> *tail; // pointer to next node
};

// ----------------------------------------------------------------------------
//
// a bag of threads with results of type <T>
//
// ----------------------------------------------------------------------------
template <class T> 
class ThreadBagBase: public ThreadBagCore
{
  ResultList<T> *result_list;   // thread results before bag is closed
  ResultList<T> *result_tail;   // end of result list
  T *result_array;              // thread results after bag is closed
  public:
  void init(void);                     // initialize thread
  void close(void);                    // close thread
  void close(int k);                   // close threads with <k> more threads
  void wait(const int n, int *d, T *); // wait for <n>-th result
  void deliver(CoreBagThread *t, T r); // deliver result <r> as thread <t>
  void scan(void);                     // scan function for this type
};

// ----------------------------------------------------------------------------
//
// bag pointer typed in result only
//
// ----------------------------------------------------------------------------
template <class R>
class ThreadBag: public Pointer< ThreadBagBase<R> >
{
  public:
  ThreadBag(void);                        // nil thread bag
  ThreadBag(int n);                       // bag for <n> threads
  static ThreadBag<R> Open(void);         // bag for arbitrary threads
  static ThreadBag<R> Open(int n);        // bag for <n> threads
  void open(void);                        // bag for arbitrary threads
  void open(int n);                       // bag for <n> threads
  friend R Wait(ThreadBag<R> bag, int n); // wait for <n>-th result in <bag>
  friend R Wait(ThreadBag<R> bag, int n, int *done);
  friend R Size(ThreadBag<R> bag);        // definitive size of bag
};

// ----------------------------------------------------------------------------
//
// a bag thread with result of type <T>
//
// ----------------------------------------------------------------------------
template <class R> 
class ResBagThread: public CoreBagThread
{
  friend class ThreadBagBase<R>;
  protected:
  ThreadBag<R> bag;            // associated bag
  void add(ThreadBag<R> _bag); // add thread to bag
  void scan(void);             // scan function for this type
};

// ----------------------------------------------------------------------------
//
// bag thread with zero arguments
//
// ----------------------------------------------------------------------------
template <class R>
class ArgBagThread0: public ResBagThread<R>
{
  R (*f)(void);
  void user(void);
  void scan(void);
  public:
  void init(ThreadBag<R> _bag, R (*_f)(void));
};

// ----------------------------------------------------------------------------
//
// bag thread with one argument
//
// ---------------------------------------------------------------------------
template <class R, class A1>
class ArgBagThread1: public ResBagThread<R>
{
  R (*f)(A1);
  A1 a1;
  void user(void);
  void scan(void);
  public:
  void init(ThreadBag<R> _bag, R (*_f)(A1), A1 _a1);
};

// ----------------------------------------------------------------------------
//
// bag thread with two arguments
//
// ---------------------------------------------------------------------------
template <class R, class A1, class A2>
class ArgBagThread2: public ResBagThread<R>
{
  R (*f)(A1, A2);
  A1 a1;
  A2 a2;
  void user(void);
  void scan(void);
  public:
  void init(ThreadBag<R> _bag, R (*_f)(A1, A2), A1 _a1, A2 _a2);
};

// ----------------------------------------------------------------------------
//
// bag thread with three arguments
//
// ---------------------------------------------------------------------------
template <class R, class A1, class A2, class A3>
class ArgBagThread3: public ResBagThread<R>
{
  R (*f)(A1, A2, A3);
  A1 a1;
  A2 a2;
  A3 a3;
  void user(void);
  void scan(void);
  public:
  void init(ThreadBag<R> _bag, R (*_f)(A1, A2, A3), A1 _a1, A2 _a2, A3 _a3);
};

// ----------------------------------------------------------------------------
//
// bag thread with four arguments
//
// ---------------------------------------------------------------------------
template <class R, class A1, class A2, class A3, class A4>
class ArgBagThread4: public ResBagThread<R>
{
  R (*f)(A1, A2, A3, A4);
  A1 a1;
  A2 a2;
  A3 a3;
  A4 a4;
  void user(void);
  void scan(void);
  public:
  void init(ThreadBag<R> _bag, 
	     R (*_f)(A1, A2, A3, A4), A1 _a1, A2 _a2, A3 _a3, A4 _a4);
};

// ----------------------------------------------------------------------------
//
// bag thread pointer with no argument
//
// ----------------------------------------------------------------------------
template <class R>
class ThreadBag0
{
  protected:
  static ArgBagThread0<R> *init(ThreadBag<R> bag, R (*f)(void));
  public:
  static void create(ThreadBag<R> bag, R (*f)(void));
  static void start(ThreadBag<R> bag, R (*f)(void));
  ThreadBag0(ThreadBag<R> bag, R (*f)(void));
};

// ----------------------------------------------------------------------------
//
// bag thread pointer with one argument
//
// ----------------------------------------------------------------------------
template <class R, class A1>
class ThreadBag1
{
  protected:
  static ArgBagThread1<R, A1> *init(ThreadBag<R> bag, R (*f)(A1), A1 a1);
  public:
  static void create(ThreadBag<R> bag, R (*f)(A1), A1 a1);
  static void start(ThreadBag<R> bag, R (*f)(A1), A1 a1);
  ThreadBag1(ThreadBag<R> bag, R (*f)(A1), A1 a1);
};

// ----------------------------------------------------------------------------
//
// bag thread pointer with two arguments
//
// ----------------------------------------------------------------------------
template <class R, class A1, class A2>
class ThreadBag2
{
  protected:
  static ArgBagThread2<R, A1, A2> *
    init(ThreadBag<R> bag, R (*f)(A1, A2), A1 a1, A2 a2);
  public:
  static void create(ThreadBag<R> bag, R (*f)(A1, A2), A1 a1, A2 a2);
  static void start(ThreadBag<R> bag, R (*f)(A1, A2), A1 a1, A2 a2);
  ThreadBag2(ThreadBag<R> bag, R (*f)(A1, A2), A1 a1, A2 a2);
};

// ----------------------------------------------------------------------------
//
// bag thread pointer with three arguments
//
// ----------------------------------------------------------------------------
template <class R, class A1, class A2, class A3>
class ThreadBag3
{
  protected:
  static ArgBagThread3<R, A1, A2, A3> *
    init(ThreadBag<R> bag, R (*f)(A1, A2, A3), A1 a1, A2 a2, A3 a3);
  public:
  static void 
    create(ThreadBag<R> bag, R (*f)(A1, A2, A3), A1 a1, A2 a2, A3 a3);
  static void
    start(ThreadBag<R> bag, R (*f)(A1, A2, A3), A1 a1, A2 a2, A3 a3);
  ThreadBag3(ThreadBag<R> bag, R (*f)(A1, A2, A3), A1 a1, A2 a2, A3 a3);
};

// ----------------------------------------------------------------------------
//
// bag thread pointer with four arguments
//
// ----------------------------------------------------------------------------
template <class R, class A1, class A2, class A3, class A4>
class ThreadBag4
{
  protected:
  static ArgBagThread4<R, A1, A2, A3, A4> *
    init(ThreadBag<R> bag, 
	 R (*f)(A1, A2, A3, A4), A1 a1, A2 a2, A3 a3, A4 a4);
  public:
  static void 
    create(ThreadBag<R> bag, 
	   R (*f)(A1, A2, A3, A4), A1 a1, A2 a2, A3 a3, A4 a4);
  static void 
    start(ThreadBag<R> bag, 
	  R (*f)(A1, A2, A3, A4), A1 a1, A2 a2, A3 a3, A4 a4);
  ThreadBag4(ThreadBag<R> bag, 
	     R (*f)(A1, A2, A3, A4), A1 a1, A2 a2, A3 a3, A4 a4);
};

#endif
// ----------------------------------------------------------------------------
// end of rt_tbag.h
// ----------------------------------------------------------------------------
// Local Variables: ***
// mode: C++ ***
// End: ***
