// ----------------------------------------------------------------------------
//
// rt_shared.cc
// RT++ shared memory interface
//
// $Id: rt_shared.cc,v 1.21 1996/04/02 16:34:51 schreine Exp $
// (C) 1996 Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
//
// ----------------------------------------------------------------------------

#include "rt++.h"

class Shared
{
  Mutex mutex;         // mutual exclusion on memory
  void *memory;        // address of free memory area
  int sz;              // size of memory
  public:	       
  void init(int n);    // initialize shared memory
  int size(void);      // size of memory
  void *malloc(int n); // allocate shared memory
  void free(void *);   // free shared memory
};

static void *smin;     // lower bound of allocated memory area
static void *smax;     // upper bound of allocated memory area
static Shared *shared; // shared object with automatic shutdown

// ----------------------------------------------------------------------------
//
// initialize shared memory system with <n> bytes
// and allocate object <shared> in shared memory
//
// ----------------------------------------------------------------------------
void 
shm_init(int n)
{
  smin = rt_shm_alloc(n); // allocate shared memory
  smax = smin+n;          // upper bound of memory
  shared = (Shared*)smin; // map shared memory object
  shared->init(n);        // initialize shared memory object
}

// ----------------------------------------------------------------------------
//
// shutdown shared memory
//
// ----------------------------------------------------------------------------
void 
shm_shut(void)
{
  rt_shm_free();
}

// ----------------------------------------------------------------------------
//
// size of shared memory
//
// ----------------------------------------------------------------------------
int
shm_size(void)
{
  return(shared->size());
}

// ----------------------------------------------------------------------------
//
// initialize memory object
//
// ----------------------------------------------------------------------------
void
Shared::init(int _sz)
{
  mutex.init();                   // initialize mutex
  memory = smin + sizeof(Shared); // address of free memory area
  sz = _sz;                       // size of memory                  
}

// ----------------------------------------------------------------------------
//
// size of memory
//
// ----------------------------------------------------------------------------
int
Shared::size(void)
{
  return(sz);
}

// ----------------------------------------------------------------------------
//
// allocate <n> bytes of shared memory suitably aligned
// return 0 if memory cannot be allocated
//
// ----------------------------------------------------------------------------
inline void *
Shared::malloc(int n)
{
  mutex.lock();                        // compute aligned memory address
  void *p = rt_align(memory);          // aligned pointer
  memory = p+n;                        // address of free memory area
  if (memory <= smax)                  // no memory overflow
    {
      mutex.unlock();                  // unlock shared memory
      return(p);                       // return pointer
    }
  else
    {
      mutex.unlock();                  // unlock shared memory 
      return(0);                       // return null pointer
    }                           
}

// ----------------------------------------------------------------------------
//
// free allocated shared memory at address <p>
//
// ----------------------------------------------------------------------------
inline void
Shared::free(void *)
{
}

// ----------------------------------------------------------------------------
//
// allocate shared memory
// abort if memory not available
//
// ----------------------------------------------------------------------------
void*
shm_alloc(int n) 
{
  void *p = shared->malloc(n);
  if (!p) rt_abort("out of memory");
  return(p);
}

// ----------------------------------------------------------------------------
//
// allocate shared memory
// return 0 if memory not available
//
// ----------------------------------------------------------------------------
void*
shm_alloc_test(int n) 
{
  return(shared->malloc(n));
}

// ----------------------------------------------------------------------------
//
// free shared memory
//
// ----------------------------------------------------------------------------
void 
shm_free(void *p) 
{
  shared->free(p);
}

// ----------------------------------------------------------------------------
// end of rt_shared.cc
// ----------------------------------------------------------------------------
// Local Variables: ***
// mode: C++ ***
// End: ***
