// ----------------------------------------------------------------------------
//
// rt_system.cc
// RT++ global system data
//
// $Id: rt_system.cc,v 1.37 1996/04/02 07:34:28 schreine Exp $
// (C) 1996 Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
//
// ----------------------------------------------------------------------------

#include "rt++.h"

// system-global data, dynamically allocated in shared memory
System *sys = 0; 

// ----------------------------------------------------------------------------
// 
// allocate global data in shared memory
//
// ----------------------------------------------------------------------------
void 
sys_init(int procnum, int stacksize) 
{
  sys = (System*)shm_alloc(sizeof(System));
  sys->System(procnum, stacksize);
}

// ----------------------------------------------------------------------------
//
// dummy function for polling
//
// ----------------------------------------------------------------------------
static volatile int dummy(int x)
{
  return(x);
}

// ----------------------------------------------------------------------------
//
// return approximative number of loop iterations with function calls 
// that fit into 100 ms (= 10% of an alarm(1) interval)
//
// ----------------------------------------------------------------------------
static int calibrate(void)
{
  int s = 0;
  const int n = 1000000;
  int t = real_clock();
  for (int i = 0; i < n; i++) 
    {
      s += dummy(i);
    }
  t = ms_clock(real_clock()-t);
  if (t == 0) t = 10;
  return(n*100/t);
}

// ----------------------------------------------------------------------------
//
// initialize system object
//
// ----------------------------------------------------------------------------
System::System(int n, int s): 
ready_hd(0), ready_tl(0), lazy(0), root(0), size(s), 
procnumber(n), regnumber(0), code(-1), stacks(0), pollcnt(calibrate())
{ 
  procid = (pid_t*)shm_alloc(procnumber*sizeof(pid_t));
  bound_hd = (CoreThread**)shm_alloc(procnumber*sizeof(CoreThread *));
  bound_tl = (CoreThread**)shm_alloc(procnumber*sizeof(CoreThread *));
  mutex_bound = (Mutex*)shm_alloc(procnumber*sizeof(Mutex));
  for (int i = 0; i < procnumber; i++)
    {
      bound_hd[i] = 0;
      bound_tl[i] = 0;
      mutex_bound[i].init();
    }
}

// ----------------------------------------------------------------------------
//
// return processor number
//
// ----------------------------------------------------------------------------
int
System::procnum(void)
{ 
  return(procnumber);
}

// ----------------------------------------------------------------------------
//
// return default stack size
//
// ----------------------------------------------------------------------------
int
System::stacksize(void)
{ 
  return(size);
}

// ----------------------------------------------------------------------------
// 
// register Unix process identifier <pid> at position <i>
//
// ----------------------------------------------------------------------------
void
System::proc(int i, pid_t pid) 
{
  if (i >= procnumber) rt_abort("process index too large");
  procid[i] = pid;
}

// ----------------------------------------------------------------------------
//
// place thread into bound queue <n>
// caller is in charge of locking processor!!!
//
// ----------------------------------------------------------------------------
inline void 
System::is_bound(CoreThread *next)
{
  const int n = next->binding();
  mutex_bound[n].lock();      // lock bound queue
  next->next = 0;             // last thread in queue
  if (bound_tl[n])            // there was a previous thread in queue
    bound_tl[n]->next = next; // link previous thread with thread
  else                        // no previous thread in queue
    bound_hd[n] = next;       // this is the first thread in queue
  bound_tl[n] = next;         // this is the last thread in queue
  mutex_bound[n].unlock();    // unlock bound queue
}

// ----------------------------------------------------------------------------
//
// get thread from binding queue of current processor
// caller is in charge of locking the processor!!!
//
// ----------------------------------------------------------------------------
inline CoreThread*
System::get_bound(void)
{
  const int n = processor.number();      // current processor
  if (!bound_hd[n]) return(0);           // cheap test first
  mutex_bound[n].lock();                 // lock bound queue
  CoreThread *next = bound_hd[n];        // first thread in queue
  while (next && (next->is_stopped()))
    {
      CoreThread *thread = next;
      next = next->next;
      thread->destroy();                 // deallocate its stack
    }
  if (next)                              // there was a thread in queue
    {
      bound_hd[n] = next->next;          // next thread in queue
      if (!bound_hd[n]) bound_tl[n] = 0; // queue is empty
    }
  mutex_bound[n].unlock();               // unlock bound queue
  return(next);                          // return thread
}

// ----------------------------------------------------------------------------
//
// place thread into ready queue
// caller is in charge of locking processor!!!
//
// ----------------------------------------------------------------------------
void 
System::is_ready(CoreThread *next)
{
  int n = next->binding(); // get thread binding
  if (n >= 0)              // thread has a binding
    {
      is_bound(next);      // place thread into specific queue
      return;
    }
  mutex_ready.lock();      // lock ready queue
  next->next = 0;          // last thread in queue
  if (ready_tl)            // there was a previous thread in queue
    ready_tl->next = next; // link previous thread with thread
  else                     // no previous thread in queue
    ready_hd = next;       // this is the first thread in queue
  ready_tl = next;         // this is the last thread in queue
  mutex_ready.unlock();	   // unlock ready queue
}

// ----------------------------------------------------------------------------
//
// get thread from ready queue
// caller is in charge of locking processor!!!
//
// ----------------------------------------------------------------------------
CoreThread*
System::get_ready(void)
{
  CoreThread *thread = get_bound(); // try bound thread first
  if (thread) return(thread);
  if (!ready_hd) return(0);         // cheap test first
  mutex_ready.lock();               // lock ready queue
  CoreThread *next = ready_hd;      // first thread in queue
  while (next && (next->is_stopped()))
    {
      CoreThread *thread = next;
      next = next->next;
      thread->destroy();           // deallocate its stack
    }
  if (next)                        // there was a thread in queue
    {
      ready_hd = next->next;       // next thread in queue
      if (!ready_hd) ready_tl = 0; // queue is empty
    }
  mutex_ready.unlock();            // unlock ready queue
  return(next);                    // return thread
}

// ----------------------------------------------------------------------------
//
// place thread into lazy list
// caller is in charge of locking processor!!!
//
// ----------------------------------------------------------------------------
void 
System::is_lazy(CoreThread *next)
{
  mutex_lazy.lock();           // lock lazy list
  if (lazy) lazy->prev = next; // old first thread in list
  next->next = lazy;           // link to old first thread
  next->prev = 0;              // no previous thread
  lazy = next;                 // this is the first thread
  mutex_lazy.unlock();         // unlock lazy list
}

// ----------------------------------------------------------------------------
//
// get thread from lazy list
// caller is in charge of locking processor!!!
//
// ----------------------------------------------------------------------------
CoreThread *
System::get_lazy(void)
{
  CoreThread *next;
  do {
    mutex_lazy.lock();               // lock lazy 
    next = lazy;                     // first thread in list
    if (next)                        // there is a thread in list
      {
	lazy = next->next;           // new lazy list
	if (lazy) lazy->prev = 0;    // no previous thread
	next->next = 0;              // unlink thread
      }
    mutex_lazy.unlock();             // unlock lazy list
  } while(next && !next->execute()); // test and set flag
  return(next);                      // return thread
}

// ----------------------------------------------------------------------------
//
// get denoted thread from lazy list
// caller is in charge of locking processor!!!
//
// ----------------------------------------------------------------------------
void
System::get_lazy(CoreThread *thread)
{
  mutex_lazy.lock();		     // lock lazy list
  CoreThread *prev = thread->prev;   // predecessor thread
  CoreThread *next = thread->next;   // successor thread
  if (prev)                          // there is a predecessor
    {
      prev->next = next;             // link to successor
      thread->prev = 0;              // unlink thread
    }
  if (next)                          // there is a successor
    {
      next->prev = prev;             // link to predecessor
      thread->next = 0;              // unlink thread
    }
  if (lazy == thread) lazy = next;   // new lazy list
  mutex_lazy.unlock();               // unlock lazy list
}

// ----------------------------------------------------------------------------
//
// lock system for termination
//
// ----------------------------------------------------------------------------
void
System::lock(void)
{ 
  mutex_term.lock();
}

// ----------------------------------------------------------------------------
//
// trigger program shutdown
//
// ----------------------------------------------------------------------------
void
System::shutdown(int c)
{
  code = c;                  // program return code
  kill (procid[0], SIGHUP);  // signal parent process (maybe myself)
  while(1) {};               // wait for termination
}

// ----------------------------------------------------------------------------
//
// terminate children processes
// must be only called by the initial process
//
// ----------------------------------------------------------------------------
int
System::terminate(void)
{
  processor.lock();                    // prevent preemption
  signal(SIGCHLD, SIG_IGN);            // ignore death of children
  for (int i = 1; i < procnumber; i++) // traverse all processes
    {
      kill(procid[i], SIGHUP);         // terminate child process
      waitpid(procid[i], 0, 0);        // wait for termination
    }
  return(code);                        // goto non-local exit
}

// ----------------------------------------------------------------------------
//
// return return code
//
// ----------------------------------------------------------------------------
int
System::ret_code(void)
{
  return(code);
}

// ----------------------------------------------------------------------------
//
// return true iff thread available
//
// ----------------------------------------------------------------------------
int
System::thread_avail(void)
{
  return(ready_hd != 0 || lazy != 0 || bound_hd[processor.number()] != 0);
}

// ----------------------------------------------------------------------------
//
// add thread to root set of garbage collection
// caller is in charge of locking processor!!!
//
// ----------------------------------------------------------------------------
void
System::root_thread(CoreThread *thread)
{
  mutex_root.lock();
  thread->root_next = root;
  thread->root_prev = 0;
  root = thread;
  mutex_root.unlock();
}

// ----------------------------------------------------------------------------
//
// remove thread from root set of garbage collection
// caller is in charge of locking processor!!!
//
// ----------------------------------------------------------------------------
void
System::no_root_thread(CoreThread *thread)
{
  mutex_root.lock();
  CoreThread *next = thread->root_next;
  CoreThread *prev = thread->root_prev;
  if (next)
    {
      next->root_prev = prev;
    }
  if (prev)
    {
      prev->root_next = next;
    }
  if (thread == root)
    {
      root = next;
    }
  mutex_root.unlock();
}

// ----------------------------------------------------------------------------
//
// return root set
//
// ----------------------------------------------------------------------------
CoreThread *
System::root_threads(void)
{
  return(root);
}

// ----------------------------------------------------------------------------
//
// filter unmarked threads from thread list
// if <stack> is set, the threads are real and are destroyed
//
// ----------------------------------------------------------------------------
static void
marked(CoreThread **head, CoreThread **tail, int stack)
{
  CoreThread *tlist = *head;
  CoreThread *prev = 0;
  CoreThread *next = tlist;
  while (next)
    {
      if (Block::block(next)->is_marked())
	{
	  prev = next;
	  next = next->next;
	  continue;
	}
      if (stack) next->destroy();
      next = next->next;
      if (prev)
	prev->next = next;
      else
	tlist = next;
    }
  *head = tlist;
  if (tail) *tail = prev;
}

// ----------------------------------------------------------------------------
//
// set lazy list
//
// ----------------------------------------------------------------------------
void
System::gc_threads(void)
{
  marked(&lazy, 0, 0);
  marked(&ready_hd, &ready_tl, 1);
  for (int i = 0; i < procnumber; i++)
    {
      marked(&bound_hd[i], &bound_tl[i], 1);
    }
}

// ----------------------------------------------------------------------------
//
// set signal value
//
// ----------------------------------------------------------------------------
void
System::set(int s)
{
  sigval = s;
}

// ----------------------------------------------------------------------------
//
// get signal value
//
// ----------------------------------------------------------------------------
int
System::get(void)
{
  return(sigval);
}

// ----------------------------------------------------------------------------
//
// return polling counter
//
// ----------------------------------------------------------------------------
int
System::poll(void)
{
  return(pollcnt);
}

// ----------------------------------------------------------------------------
//
// trigger garbage collection
//
// ----------------------------------------------------------------------------
void
System::gc(void)
{
  int number = processor.number();
  for (int i = 0; i < procnumber; i++)
    {
      if (i != number)
	{
	  int p = procid[i];
	  do
	    {
	      set(0);
	      kill(p, SIGUSR1);
	      while (get() == 0) { }
	    }
	  while (get() == 1);
	}
    }
}

// ----------------------------------------------------------------------------
// 
// global synchronization
//
// ----------------------------------------------------------------------------
void
System::wait(void)
{
  mutex_gc.lock();
  regnumber++;
  mutex_gc.unlock();
  while (regnumber < procnumber) { }
}

// ----------------------------------------------------------------------------
// 
// allocate new stack
//
// ----------------------------------------------------------------------------
inline void *
System::stack(int s)
{

  mutex_stacks.lock();
  void *st = stacks;
  if (st)
    {
      stacks = *(void**)st;
      mutex_stacks.unlock();
    }
  else
    {
      mutex_stacks.unlock();
      st = shm_alloc(s);
    }
  return(st);
}

// ----------------------------------------------------------------------------
// 
// allocate new stack
//
// ----------------------------------------------------------------------------
void *
newstack(int s)
{
  return(sys->stack(s));
}

// ----------------------------------------------------------------------------
// 
// dispose stack
// caller is in charge of locking the processor!!!
//
// ----------------------------------------------------------------------------
void
System::dispose(void *st)
{
  mutex_stacks.lock();
  *(void**)st = stacks;
  stacks = st;
  mutex_stacks.unlock();
}

// ----------------------------------------------------------------------------
// end of rt_system.cc
// ----------------------------------------------------------------------------
// Local Variables: ***
// mode: C++ ***
// End: ***
